@@ -1,27 +1,21 @@
 // Copyright Maarten L. Hekkelman, Radboud University 2008-2013.
-//        Copyright Maarten L. Hekkelman, 2014-2020
+//        Copyright Maarten L. Hekkelman, 2014-2022
 // Distributed under the Boost Software License, Version 1.0.
 //    (See accompanying file LICENSE_1_0.txt or copy at
 //          http://www.boost.org/LICENSE_1_0.txt)
 
-#include <boost/algorithm/string.hpp>
+#include <filesystem>
 
 #include <zeep/crypto.hpp>
 #include <zeep/http/server.hpp>
-#include <filesystem>
+#include <zeep/json/parser.hpp>
+#include <zeep/http/uri.hpp>
 
-namespace ba = boost::algorithm;
 namespace fs = std::filesystem;
 
 namespace zeep::http
 {
 
-namespace
-{
-// a regex for URI
-const std::regex kURIRx(R"((?:(https?:)(?://([^/]+)))?(?:/?([^?]*))(?:\?(.+))?)", std::regex_constants::icase);
-}
-
 request::request(const std::string& method, const std::string& uri, std::tuple<int,int> version,
 		std::vector<header>&& headers, std::string&& payload)
 	: m_method(method)
@@ -29,9 +23,48 @@ request::request(const std::string& method, const std::string& uri, std::tuple<i
 	, m_headers(std::move(headers))
 	, m_payload(std::move(payload))
 {
-	m_version[0] = '0' + std::get<0>(version);
+	m_version[0] = static_cast<char>('0' + std::get<0>(version));
 	m_version[1] = '.';
-	m_version[2] = '0' + std::get<1>(version);
+	m_version[2] = static_cast<char>('0' + std::get<1>(version));
+}
+
+request::request(const request& req)
+	: m_local_address(req.m_local_address)
+	, m_local_port(req.m_local_port)
+	, m_method(req.m_method)
+	, m_uri(req.m_uri)
+	, m_headers(req.m_headers)
+	, m_payload(req.m_payload)
+	, m_close(req.m_close)
+	, m_timestamp(req.m_timestamp)
+	, m_credentials(req.m_credentials)
+	, m_remote_address(req.m_remote_address)
+{
+	char* d = &m_version[0];
+	for (auto c: req.m_version)
+		*d++ = c;
+}
+
+request& request::operator=(const request& req)
+{
+	if (this != &req)
+	{
+		m_local_address = req.m_local_address;
+		m_local_port = req.m_local_port;
+		m_method = req.m_method;
+		m_uri = req.m_uri;
+		m_version[0] = req.m_version[0];
+		m_version[1] = req.m_version[1];
+		m_version[2] = req.m_version[2];
+		m_headers = req.m_headers;
+		m_payload = req.m_payload;
+		m_close = req.m_close;
+		m_timestamp = req.m_timestamp;
+		m_credentials = req.m_credentials;
+		m_remote_address = req.m_remote_address;
+	}
+
+	return *this;
 }
 
 void request::set_local_endpoint(boost::asio::ip::tcp::socket& socket)
@@ -40,30 +73,6 @@ void request::set_local_endpoint(boost::asio::ip::tcp::socket& socket)
 	m_local_port = socket.local_endpoint().port();
 }
 
-std::string request::get_path() const
-{
-	std::smatch m;
-	if (not std::regex_match(m_uri, m, kURIRx))
-		throw std::invalid_argument("the request uri is not valid");
-	return "/" + fs::path(m[3]).lexically_normal().string();
-}
-
-std::string request::get_query() const
-{
-	std::smatch m;
-	if (not std::regex_match(m_uri, m, kURIRx))
-		throw std::invalid_argument("the request uri is not valid");
-	return m[4];
-}
-
-std::string request::get_host() const
-{
-	std::smatch m;
-	if (not std::regex_match(m_uri, m, kURIRx))
-		throw std::invalid_argument("the request uri is not valid");
-	return m[2];
-}
-
 float request::get_accept(const char* type) const
 {
 	float result = 1.0f;
@@ -190,7 +199,7 @@ void request::remove_header(const char* name)
 		m_headers.end());
 }
 
-std::pair<std::string,bool> get_urlencode_parameter(const std::string& s, const char* name)
+std::pair<std::string,bool> get_urlencoded_parameter(const std::string& s, const char* name)
 {
 	std::string::size_type b = 0;
 	std::string result;
@@ -229,22 +238,40 @@ std::tuple<std::string,bool> request::get_parameter_ex(const char* name) const
 	std::string result, contentType = get_header("Content-Type");
 	bool found = false;
 
-	if (ba::starts_with(contentType, "application/x-www-form-urlencoded"))
+	if (starts_with(contentType, "application/x-www-form-urlencoded"))
 	{
-		tie(result, found) = get_urlencode_parameter(m_payload, name);
+		tie(result, found) = get_urlencoded_parameter(m_payload, name);
 		if (found)
 			return std::make_tuple(result, true);
 	}
 
-	auto b = m_uri.find('?');
-	if (b != std::string::npos)
+	uri uri(m_uri);
+	auto query = uri.get_query();
+
+	if (not query.empty())
 	{
-		tie(result, found) = get_urlencode_parameter(m_uri.substr(b + 1), name);
+		tie(result, found) = get_urlencoded_parameter(query, name);
 		if (found)
 			return std::make_tuple(result, true);
 	}
 
-	if (ba::starts_with(contentType, "multipart/form-data"))
+	if (starts_with(contentType, "application/json"))
+	{
+		try
+		{
+			json::element e;
+			json::parse_json(m_payload, e);
+			if (e.is_object() and e.contains(name))
+			{
+				result = e.at(name).as<std::string>();
+				found = true;
+			}
+		}
+		catch (const std::exception &)
+		{
+		}
+	}
+	else if (starts_with(contentType, "multipart/form-data"))
 	{
 		std::string::size_type b = contentType.find("boundary=");
 		if (b != std::string::npos)
@@ -315,7 +342,7 @@ std::tuple<std::string,bool> request::get_parameter_ex(const char* name) const
 			}
 		}
 
-		ba::replace_all(result, "\r\n", "\n");
+		replace_all(result, "\r\n", "\n");
 	}
 	
 	return make_tuple(result, found);
@@ -329,14 +356,13 @@ std::multimap<std::string,std::string> request::get_parameters() const
 	{
 		std::string contentType = get_header("Content-Type");
 		
-		if (ba::starts_with(contentType, "application/x-www-form-urlencoded"))
+		if (starts_with(contentType, "application/x-www-form-urlencoded"))
 			ps = m_payload;
 	}
 	else if (m_method == "GET" or m_method == "PUT")
 	{
-		std::string::size_type d = m_uri.find('?');
-		if (d != std::string::npos)
-			ps = m_uri.substr(d + 1);
+		uri uri(m_uri);
+		ps = uri.get_query();
 	}
 
 	std::multimap<std::string,std::string> parameters;
@@ -384,7 +410,7 @@ std::multimap<std::string,std::string> request::get_parameters() const
 	// 			{
 	// 				std::string contentType = get_header("Content-Type");
 					
-	// 				if (ba::starts_with(contentType, "application/x-www-form-urlencoded"))
+	// 				if (starts_with(contentType, "application/x-www-form-urlencoded"))
 	// 					ps = m_payload;
 	// 			}
 	// 			break;
@@ -456,7 +482,7 @@ file_param_parser::file_param_parser(const request& req, const std::string& payl
 {
 	std::string contentType = m_req.get_header("Content-Type");
 
-	if (ba::starts_with(contentType, "multipart/form-data"))
+	if (starts_with(contentType, "multipart/form-data"))
 	{
 		std::string::size_type b = contentType.find("boundary=");
 		if (b != std::string::npos)
@@ -533,11 +559,11 @@ file_param file_param_parser::next()
 
 				auto b = p.begin();
 				auto e = p.end();
-				std::match_results<std::string::iterator> m;
-				while (b < e and std::regex_search(b, e, m, re))
+				std::match_results<std::string::iterator> m2;
+				while (b < e and std::regex_search(b, e, m2, re))
 				{
-					auto key = m[1].str();
-					auto value = m[2].str();
+					auto key = m2[1].str();
+					auto value = m2[2].str();
 					if (value.length() > 1 and ((value.front() == '"' and value.back() == '"') or (value.front() == '\'' and value.back() == '\'')))
 						value = value.substr(1, value.length() - 2);
 
@@ -546,13 +572,13 @@ file_param file_param_parser::next()
 					else if (key == "filename")
 						result.filename = value;
 
-					b = m[0].second;
+					b = m2[0].second;
 				}
 			}
 			else if (std::regex_match(m_payload.begin() + l, m_payload.begin() + m_i, m, k_rx_cont))
 			{
 				result.mimetype = m[1].str();
-				if (ba::starts_with(result.mimetype, "multipart/"))
+				if (starts_with(result.mimetype, "multipart/"))
 					throw std::runtime_error("multipart file uploads are not supported");
 			}
 		}
@@ -599,11 +625,11 @@ std::string request::get_cookie(const char* name) const
 			continue;
 
 		std::vector<std::string> rawCookies;
-		ba::split(rawCookies, h.value, ba::is_any_of(";"));
+		split(rawCookies, h.value, ";");
 
 		for (std::string& cookie : rawCookies)
 		{
-			ba::trim(cookie);
+			trim(cookie);
 
 			auto d = cookie.find('=');
 			if (d == std::string::npos)
@@ -626,11 +652,11 @@ void request::set_cookie(const char* name, const std::string& value)
 			continue;
 		
 		std::vector<std::string> rawCookies;
-		ba::split(rawCookies, h.value, ba::is_any_of(";"));
+		split(rawCookies, h.value, ";");
 
 		for (std::string& cookie : rawCookies)
 		{
-			ba::trim(cookie);
+			trim(cookie);
 
 			auto d = cookie.find('=');
 			if (d == std::string::npos)
@@ -713,7 +739,7 @@ std::locale& request::get_locale() const
 
 		std::string preferred;
 		std::vector<std::string> accepted;
-		ba::split(accepted, acceptedLanguage, ba::is_any_of(","));
+		split(accepted, acceptedLanguage, ",");
 
 		struct lang_score
 		{
@@ -739,7 +765,7 @@ std::locale& request::get_locale() const
 				if (iequals(loc.name(), name))
 					scores.push_back({lang, region, score, loc});
 			}
-			catch(const std::exception& e)
+			catch(const std::exception &)
 			{
 			}
 		};
@@ -844,7 +870,7 @@ std::ostream& operator<<(std::ostream& io, const request& req)
 	std::vector<boost::asio::const_buffer> buffers = req.to_buffers();
 
 	for (auto& b: buffers)
-		io.write(boost::asio::buffer_cast<const char*>(b), boost::asio::buffer_size(b));
+		io.write(static_cast<const char*>(b.data()), b.size());
 
 	return io;
 }
